import dataclasses
import functools
import html
import http.server
import io
import os
import posixpath
import re
import sys
import typing as tg
import urllib.parse

import argparse_subcommand as ap_sub

import base as b
import sdrl.argparser
import sdrl.constants as c
import sdrl.course
import sdrl.macros as macros
import sdrl.macroexpanders as macroexpanders
import sdrl.markdown as md
import sdrl.participant

meaning = """Specialized webserver for locally viewing contents of a student repo work directory."""


def add_arguments(subparser):
    subparser.add_argument('--port', '-p', type=int, default=8080,
                           help="webserver will listen on this port (default: 8080)")
    subparser.add_argument('--instructor', '-i', action='store_true', default=False,
                           help=f"generate task links to the instructor versions (not the student versions)")


class CourseDummy(sdrl.course.Course):
    blockmacro_topmatter = dict()
    
    def __init__(self, *args, **kwargs):
        pass  # we are a truly dumb Dummy!


def execute(pargs: ap_sub.Namespace):
    b.set_loglevel('INFO')
    b.info(f"Webserver starts. Visit 'http://localhost:{pargs.port}/'. Terminate with Ctrl-C.")
    b.set_register_files_callback(lambda s: None)
    macroexpanders.register_macros(CourseDummy())  # noqa
    server = SedrilaServer(('', pargs.port), SedrilaHTTPRequestHandler, pargs=pargs)
    server.serve_forever()


def render_markdown(info: 'Info', infile, outfile):
    markup_body = infile.read().decode()
    markup = f"# {info.lastname}:{info.fullpath}\n\n{info.byline}\n\n{markup_body}\n"
    do_render_markdown(info, markup, outfile)


def render_prot(info: 'Info', infile, outfile):
    markup = (f"# {info.lastname}:{info.fullpath}\n\n{info.byline}\n\n"
              f"[PROT::{info.fullpath}]\n")
    do_render_markdown(info, markup, outfile)


def render_sourcefile(language: str, info: 'Info', infile, outfile):
    src = infile.read().decode()
    markup = (f"# {info.title}\n\n{info.byline}\n\n"
              f"```{language}\n"
              f"{src}\n"
              f"```\n")
    do_render_markdown(info, markup, outfile)


def just_copyfile(copyfilefunc, info, infile, outfile):
    copyfilefunc(infile, outfile)


def do_render_markdown(info: 'Info', markup: str, outfile):
    template = """<!DOCTYPE HTML>
      <html>
      <head>
        <meta charset="{enc}">
        {csslinks}
        <title>{title}</title>
      </head>
      <body>
        {body}
      </body>
      </html>
    """
    macros.switch_part("viewer")
    mddict = md.render_markdown(info.fullpath, info.basename, markup, b.Mode.STUDENT, dict())
    htmltext = template.format(enc='utf8', csslinks=info.csslinks,
                               title=f"{info.lastname}:{info.basename}", body=mddict['html'])
    outfile.write(htmltext.encode())


@dataclasses.dataclass
class Info:
    pargs: ap_sub.Namespace
    basename: str
    fullpath: str
    lastname: str
    byline: str
    csslinks: str

    @property
    def title(self) -> str:
        return f"{self.lastname}:{self.basename}"


class SedrilaServer(http.server.HTTPServer):
    pargs: ap_sub.Namespace
    student_name: str = "N.N."
    lastname: str = "N.N."
    partner_name: str = ""
    course_url: str = ""
    submissionitems: dict
    submission_re: str

    def __init__(self, *args, **kwargs):
        self.server_version = f"SedrilaHTTP/{sdrl.argparser.SedrilaArgParser.get_version()}"
        self.pargs = kwargs.pop('pargs')
        try:
            student = sdrl.participant.Student()
            self.student_name = student.student_name
            mm = re.search(r"(\w+)$", self.student_name)
            self.lastname = mm.group(1) if mm else "??"
            self.partner_name = student.partner_student_name
            self.course_url = student.course_url
        except b.CritialError:
            pass  # fall back to defaults
        if os.path.exists(c.SUBMISSION_FILE):
            self.submissionitems = b.slurp_yaml(c.SUBMISSION_FILE)
            self.submission_re = self._matcher_regexp(self.submissionitems.keys())
        else:
            self.submissionitems = dict()
            self.submission_re = None
        super().__init__(*args, **kwargs)

    def version_string(self) -> str:
        return self.server_version
    
    def _matcher_regexp(self, items: tg.Iterable[str]) -> str:
        return '|'.join([re.escape(item) for item in items])

class SedrilaHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    """Serve nice directory listings, serve rendered versions of some file types and files as-is otherwise."""
    server: SedrilaServer
    renderer: tg.Callable[[tg.Any, tg.Any], None]  # for-read() file, for-write() file
    extensions_map = _encodings_map_default = {
        '.gz': 'application/gzip',
        '.Z': 'application/octet-stream',
        '.bz2': 'application/x-bzip2',
        '.xz': 'application/x-xz',
    }
    how_to_render = dict(  # suffix -> (mimetype, renderfunc)
        md=('text/html', render_markdown),
        c=('text/html', functools.partial(render_sourcefile, 'c')),
        css=('text/html', functools.partial(render_sourcefile, 'css')),
        html=('text/html', functools.partial(render_sourcefile, 'html')),
        json=('text/html', functools.partial(render_sourcefile, 'json')),
        java=('text/html', functools.partial(render_sourcefile, 'java')),
        py=('text/html', functools.partial(render_sourcefile, 'python')),
        sh=('text/html', functools.partial(render_sourcefile, 'sh')),
        yaml=('text/html', functools.partial(render_sourcefile, 'yaml')),
        prot=('text/html', render_prot),
    )

    def send_head(self):  # simplified: no caching, no Content-Length
        path = self.translate_path(self.path)
        f = None  # noqa
        if os.path.isdir(path):
            parts = urllib.parse.urlsplit(self.path)
            if not parts.path.endswith('/'):
                self.send_response(http.HTTPStatus.MOVED_PERMANENTLY)
                new_parts = (parts[0], parts[1], parts[2] + '/',
                             parts[3], parts[4])
                new_url = urllib.parse.urlunsplit(new_parts)
                self.send_header("Location", new_url)
                self.send_header("Content-Length", "0")
                self.end_headers()
                return None
            return self.list_directory(path)
        ctype = self.guess_type(path)
        if path.endswith("/"):
            self.send_error(http.HTTPStatus.NOT_FOUND, "File not found")
            return None
        try:
            f = open(self.sedrila_actualpath(path), 'rb')
        except (OSError, FileNotFoundError):
            self.send_error(http.HTTPStatus.NOT_FOUND, "File not found")
            return None
        try:
            self.send_response(http.HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.end_headers()
            return f
        except:
            f.close()
            raise

    def list_directory(self, path):
        try:
            dirlist = os.listdir(path)
        except OSError:
            self.send_error(
                http.HTTPStatus.NOT_FOUND,
                "No permission to list directory")
            return None
        dirlist.sort(key=lambda a: a.lower())
        pairslist = [(name, os.path.join(path, name)) for name in dirlist]
        dirpairs = [(name, fullname) for name, fullname in pairslist 
                    if os.path.isdir(fullname)]
        filepairs = [(name, fullname) for name, fullname in pairslist 
                     if os.path.isfile(fullname)]
        r = []
        try:
            displaypath = urllib.parse.unquote(self.path,
                                               errors='surrogatepass')
        except UnicodeDecodeError:
            displaypath = urllib.parse.unquote(path)
        displaypath = html.escape(displaypath, quote=False)
        enc = sys.getfilesystemencoding()
        info = Info(pargs=self.server.pargs,
                    basename=displaypath, fullpath=displaypath, 
                    lastname=self.server.lastname, byline=self.sedrila_byline(),
                    csslinks=self.sedrila_csslinks())
        r.append('<!DOCTYPE HTML>')
        r.append('<html>')
        r.append('<head>')
        r.append(f'<meta charset="{enc}">')
        r.append(info.csslinks)
        r.append(f'<title>{info.title}</title>\n</head>')
        r.append(f'<body>\n<h1>{info.title}</h1>')
        r.append(f'<p>{info.byline}</p>')
        if filepairs:
            r.append('<hr>\n<h3>Files</h3>\n<ol>')
            r.extend(self.sedrila_linkitems(filepairs, info))
            r.append('</ol>')
        if dirpairs:
            r.append('<hr>\n<h3>Directories</h3>\n<ol>')
            r.extend(self.sedrila_linkitems(dirpairs, info, dirs=True))
            r.append('</ol>')
        r.append('</body>\n</html>\n')
        encoded = '\n'.join(r).encode(enc, 'surrogateescape')
        f = io.BytesIO()
        f.write(encoded)
        f.seek(0)
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "text/html; charset=%s" % enc)
        self.send_header("Content-Length", str(len(encoded)))
        self.end_headers()
        self.renderer = functools.partial(just_copyfile, super().copyfile, info)
        return f

    def copyfile(self, source, outputfile):
        """Render file-like object source into file-like destination outputfile."""
        self.renderer(source, outputfile)

    def guess_type(self, path):
        base, ext = posixpath.splitext(path)
        basename = os.path.basename(path)
        info = Info(pargs=self.server.pargs,
                    basename=basename, fullpath=os.path.relpath(path), 
                    lastname=self.server.lastname, byline=self.sedrila_byline(),
                    csslinks=self.sedrila_csslinks())
        if ext and ext[1:] in self.how_to_render:
            mimetype, renderfunc = self.how_to_render[ext[1:]]  # lookup without the dot
            self.renderer = functools.partial(renderfunc, info)
            return mimetype
        elif ext == ".htmlpage":  # special case
            self.renderer = functools.partial(just_copyfile, super().copyfile, info)
            return 'text/html'
        else:
            self.renderer = functools.partial(just_copyfile, super().copyfile, info)
        return super().guess_type(path)

    def sedrila_actualpath(self, path: str) -> str:
        """Convert *.htmlpage to *.html (special case)."""
        actual = re.sub(r"\.htmlpage$", ".html", path)
        # print(f"actualpath: {path} -> {actual}")
        return actual

    def sedrila_byline(self) -> str:
        partner = f" (and {self.server.partner_name})" if self.server.partner_name else ""
        return f"{self.server.student_name}{partner}"

    def sedrila_csslinks(self) -> str:
        if not self.server.course_url:
            return ""
        return (f'<link href="{self.server.course_url}/sedrila.css" rel="stylesheet">\n'
                f'<link href="{self.server.course_url}/local.css" rel="stylesheet">\n'
                f'<link href="{self.server.course_url}/codehilite.css" rel="stylesheet">\n')
    
    def sedrila_linkitems(self, pairs: tg.Iterable[tuple[str, str]], info: Info, dirs=False) -> tg.Iterable[str]:
        res = []
        submission_re = self.server.submission_re
        for name, fullname in pairs:
            submission_mm = submission_re and re.match(submission_re, name)
            tasklink = ""
            if self.is_sedrila_invisible(name) or os.path.islink(fullname):
                continue  # skip dotfiles and symlinks
            if not dirs and name.endswith('.html'):
                htmlpagelink = f"&#9;&#9;&#9;<a href='{name}page' class='vwr-htmlpagelink'>page</a>"
            else:
                htmlpagelink = ""
            if dirs:
                cssclass = 'vwr-dirlink'
            elif submission_mm:
                cssclass = 'vwr-filelink-submission'
                matchtext = submission_mm.group()
                instructordir = (f"{c.AUTHOR_OUTPUT_INSTRUCTORS_DEFAULT_SUBDIR}/" 
                                 if info.pargs.instructor else "")
                taskurl = f"{self.server.course_url}/{instructordir}{matchtext}.html"
                tasklink = f"&#9;&#9;&#9;<a href='{taskurl}' class='vwr-tasklink'>Task '{matchtext}'</a>"
            else:
                cssclass = 'vwr-filelink'
            href = urllib.parse.quote(name, errors='surrogatepass')
            linktext = html.escape(name, quote=False)
            res.append(f"  <li class='vwr-pre'><a href='{href}' class='{cssclass}'>{linktext}</a>"
                       f"{tasklink}{htmlpagelink}</li>")
        return res

    @staticmethod
    def is_sedrila_invisible(name: str) -> bool:
        return name.startswith('.')  # TODO 2: use https://pypi.org/project/gitignore-parser/



import functools
import json
import os

import requests

import base as b
import sdrl.constants as c


class Student:
    root: str  # where PARTICIPANT_FILE lives: '.' or '..'
    course_url: str  # SeDriLa homepage URL minus the '/index.html' part
    student_name: str
    student_id: str
    partner_student_name: str
    partner_student_id: str
    metadata_url: str  # where to get JSON config

    def __init__(self):
        self.root = "."
        try:
            # grant students some slack. if they are calling from inside a working dir, it's fine.
            if not os.path.isfile(c.PARTICIPANT_FILE) and os.path.isfile(os.path.join("..", c.PARTICIPANT_FILE)):
                self.root = ".."
            data = b.slurp_yaml(os.path.join(self.root, c.PARTICIPANT_FILE))
        except FileNotFoundError:
            b.critical(f"cannot read '{c.PARTICIPANT_FILE}'")
        try:
            self.course_url = data['course_url']  # noqa
            assert isinstance(self.course_url, str)
            self.student_name = data['student_name']
            assert isinstance(self.student_name, str)
            self.student_id = str(data['student_id'])
            self.partner_student_name = data['partner_student_name']
            assert isinstance(self.partner_student_name, str)
            self.partner_student_id = str(data['partner_student_id'])
        except KeyError:
            b.critical(f"malformed file '{c.PARTICIPANT_FILE}': must contain "
                       "course_url, student_name, student_id, partner_student_name, partner_student_id.")
        self.metadata_url = f"{self.course_url}/{c.METADATA_FILE}"

    @functools.cached_property
    def metadatadict(self) -> b.StrAnyDict:
        return self.get_metadata(self.course_url)
        
    @staticmethod    
    def get_metadata(course_url: str) -> b.StrAnyDict:
        FILE_URL_PREFIX = 'file://'
        url = os.path.join(course_url, c.METADATA_FILE)
        try:
            if url.startswith(FILE_URL_PREFIX):
                jsontext = b.slurp(url[len(FILE_URL_PREFIX):])
            else:
                jsontext = requests.get(url=url).text
        except Exception as exc:  # noqa
            jsontext = ""
            b.critical(f"Error fetching URL '{url}'.")
        try:
            metadata = json.loads(jsontext)
        except:  # noqa
            metadata = dict()
            b.critical(f"JSON format error at '{url}'")
        return metadata

    @staticmethod
    def prompts(pdict):
        pdict.setdefault('student_name', "Name")
        pdict.setdefault('student_id', "ID")
        pdict.setdefault('partner_student_name', "Partner's name")
        pdict.setdefault('partner_student_id', "Partner's ID")
        return pdict



#!/usr/bin/env python3
import os
import sys

import base
import sdrl.constants as c
import sdrl.argparser
import sdrl.subcmd  # this is where the subcommands will be found


def main():  # uses sys.argv
    """Calls subcommand given on command line"""
    parser = sdrl.argparser.SedrilaArgParser(description="-")  # description is set lazily
    parser.scan("sdrl.subcmd.*")
    if os.environ.get(c.SEDRILA_COMMAND_ENV) and (len(sys.argv) < 2 or sys.argv[1].startswith("--")):
        print(f"Using existing environment var {c.SEDRILA_COMMAND_ENV}")
        input = os.environ.get(c.SEDRILA_COMMAND_ENV)
        if len(sys.argv) >= 2:
            input += " ".join(sys.argv[1:])
        args = parser.parse_args(input.split(" "))
    else:
        args = parser.parse_args()
    try:
        parser.execute_subcommand(args)
    except base.CritialError:
        pass  # base.critical has already printed a message


if __name__ == "__main__":
    mode = 'normal'  # profile, pdb, normal
    if mode == 'profile':
        import cProfile
        cProfile.run('main()', sort='cumulative')
    elif mode == 'pdb':
        try:
            main()
        except Exception as ex:
            import pdb
            print(str(ex))
            pdb.post_mortem()
    else:
        main()  # normal life


AUTHOR_ALTDIR_KEYWORD = "ALT:"
AUTHOR_CONFIG_FILENAME = "sedrila.yaml"  # at top-level of source dir
AUTHOR_GLOSSARY_BASENAME = "glossary"  # .md at top-level of chapterdir, .html in build directory
AUTHOR_OUTPUT_INSTRUCTORS_DEFAULT_SUBDIR = "instructor"
CACHE_FILENAME = ".sedrila_cache"  # in instructor target dir
HTML_DIFFICULTY_SIGN = "&#x26ab;&#xfe0e;"  # &#x26ab; is an icon and always black, &#xfe0e; is the text-variant selector
  # https://commons.wikimedia.org/wiki/Unicode_circle_shaped_symbols
INTERACT_ACCEPT_SYMBOL = "✓"
INTERACT_REJECT_SYMBOL = "X"
HTACCESS_FILE = ".htaccess"  # in instructor part of build directory
METADATA_FILE = "course.json"  # at top-level of build directory
PARTICIPANT_FILE = "student.yaml"
REPOS_HOME_VAR = "SEDRILA_INSTRUCTOR_REPOS_HOME"
REPO_USER_CMD_VAR = "SEDRILA_INSTRUCTOR_COMMAND"
REPO_USER_CMD_DEFAULT = "/bin/bash"  # fallback only if $SHELL is not set
SUBMISSION_FILE = "submission.yaml"
SUBMISSION_COMMIT_MSG = "submission.yaml"
SUBMISSION_CHECKED_COMMIT_MSG = "submission.yaml checked"
SUBMISSION_CHECK_MARK = "CHECK"
SUBMISSION_ACCEPT_MARK = "ACCEPT"
SUBMISSION_REJECT_MARK = "REJECT"
SUBMISSION_OVERRIDE_PREFIX = "OVERRIDE_"
SUBMISSION_NONTASK_MARK = "NO_SUCH_TASKNAME"
SEDRILA_COMMAND_ENV = "SEDRILA_COMMAND"



"""
Macros extend Markdown by a [MYMACRO::arg1::arg2] syntax.
A macro expands into text and may have state.

# Architecture

## 4 macro modes

Any macro belongs to one of the following modes of working:
- EARLY: The macro produces Markdown markup and will be expanded in the preprocessor stage of 
  the Markdown module. 
  Macros of the other modes all get expanded in the postprocessor stage.
- INNER: The macro produces a fragment of HTML text within a paragraph.
- BLOCK: The macro produces a self-contained <div>...</div> block.
- BLOCKSTART: The macro produces (among other things) an HTML opening tag.
- BLOCKEND: The macro produces (possibly among other things) the corresponding closing tag.

## The <p>-problem

The most basic mechanism of Markdown is recognizing blocks of text between empty lines
as paragraphs and enclose the text in <p>...</p> tags.
This is problematic for blockstart and blockend macros, because having their opening
tag (say, <div>) between <p> and </p> creates malformed HTML.
We call this the <p>-problem.

## Solution of the <p>-problem

Let us assume [START] is a blockstart macro and [END] is a blockend macro.
In terms of its line layout, the structure of the switch from one block to the next
may look like this:

Layout 1        2           3           4

text          text        text        text
[END]         [END]
[START]                   [END]       [END]
text          [START]     [START]
              text                    [START]
                          text
                                      text

Layout 1 will become enclosed in a single <p>...</p> pair and cannot be repaired easily.
It must not be used.

Layout 2 will involve "[END]</p>" and "<p>[START]", 
which will be turned into "</p>[END]" and "[START]<p>".

Layout 3 will involve "<p>[END]\n[START]</p>", 
which will be turned into "[END]\n[START]".

Layout 4 will involve "<p>[END]</p>" and "<p>[START]</p>", 
which will be turned into "[END]" and "[START]".

This way, the tag structures created by (properly used) block macros stay intact
and the text is still properly surrounded by <p>...</p>

## Macro state

The module has a single state object called macrostate,
which is a dictionary with one slot per macro: the macro name.
Macros have a 'switcher' callback function that gets called whenever processing
of a course part starts. This can reset state if needed.

"""

import dataclasses
import enum
import itertools
import re
import typing as tg

import base as b


@dataclasses.dataclass
class Macrocall:
    """Represent where and how a macro was called, allow producing errors/warnings for it."""
    md: 'SedrilaMarkdown'  # noqa
    filename: str
    partname: str
    macrocall_text: str
    macroname: str
    arg1: b.OStr
    arg2: b.OStr

    def error(self, msg: str):
        b.error(f"{self.macrocall_text}\n   {msg}", file=self.filename)

    def warning(self, msg: str):
        b.warning(f"{self.macrocall_text}\n   {msg}", file=self.filename)


class MM(enum.Enum):
    """MacroMode"""
    EARLY = 1
    INNER = 2
    BLOCK = 3
    BLOCKSTART = 4
    BLOCKEND = 5


Macroexpander = tg.Callable[[Macrocall], str]
Partswitcher = tg.Callable[[str, str], None]  # macroname, newpartname
Macrodef = tg.Tuple[int, MM, Macroexpander, Partswitcher]  # num_args, expander, switcher

macrodefs_early: dict[str, Macrodef] = dict()  # macroname -> macrodef
macrodefs_late: dict[str, Macrodef] = dict()  # macroname -> macrodef
macrostate: dict[str, tg.Any] = dict()  # namespace -> state_object

macro_regexp = (r"(?P<ppre></?p>)?"
                r"(?P<macrocall>\[(?P<name>[A-Z][A-Z0-9_]+)(::(?P<arg1>.*?))?(::(?P<arg2>.*?))?\])"
                r"(?=[^(]|$)(?P<ppost></?p>)?")  
# bracketed all-caps: [ALL2_CAPS] with zero to two arguments: [NAME::arg] or [NAME::arg1::arg2]
# suppress matches on normal links: [TEXT](url)


def expand_macros(sourcefile: str, partname: str, markup: str, is_early_phase=False) -> str:
    """Apply matching macrodefs, report errors for non-matching macro calls."""
    def my_expand_macro(mm: re.Match) -> str:
        return expand_macro(sourcefile, partname, mm, is_early_phase)
    return re.sub(macro_regexp, my_expand_macro, markup)


def expand_macro(sourcefile: str, partname: str, mm: re.Match, is_early_phase=False) -> str:
    """Apply matching macrodef or report error or wait for late phase to report it then."""
    global macrodefs_early, macrodefs_late
    import sdrl.markdown
    macrodefs = (macrodefs_early if is_early_phase else macrodefs_late)
    ppre, call, ppost = mm.group('ppre'), mm.group('macrocall'), mm.group('ppost')
    macroname, arg1, arg2 = mm.group('name'), mm.group('arg1'), mm.group('arg2')
    macrocall = Macrocall(md=sdrl.markdown.md, filename=sourcefile, partname=partname,
                          macrocall_text=call,
                          macroname=macroname, arg1=arg1, arg2=arg2)
    my_numargs = (arg1 is not None) + (arg2 is not None)
    # ----- check name:
    if macroname not in macrodefs:
        if not is_early_phase:  # so we do not complain twice
            macrocall.error(f"Macro '{macroname}' is not defined")
        return call  # unexpanded version helps the user most
    numargs, mode, expander, switcher = macrodefs[macroname]
    # ----- check args:
    if my_numargs != numargs:
        if not is_early_phase:  # so we do not complain twice
            macrocall.error("Macro '%s' called with %d args, expects %d" %
                            (macroname, my_numargs, numargs))
        return call  # unexpanded version helps the user most
    if my_numargs > 0 and arg1 == "":
        macrocall.warning("Macro '%s' called with empty argument 1" % macroname)
    # ----- expand:
    # b.debug(f"expanding {macrocall.macrocall_text}")
    expansion = expander(macrocall)
    # ----- handle ppre and ppost:
    ppre = ppre if ppre else ""  # fill in "" for None
    ppost = ppost if ppost else ""
    if mode == MM.EARLY:
        pass  # use ppre, ppost verbatim
    elif mode == MM.INNER:
        pass  # ditto
    elif mode == MM.BLOCK:
        pass  # ditto
    elif mode == MM.BLOCKSTART:
        if not ppre and not ppost:  # Layout 1
            macrocall.warning("blockmacro blocks must be separated by empty lines _somewhere_")
        elif ppre == '<p>' and not ppost:  # Layout 2
            ppost = ppre  # shift <p> from front to back
            ppre = ""
        elif not ppre and ppost == '</p>':  # Layout 3
            ppost = ""  # remove
        elif ppre == '<p>' and ppost == '</p>':  # Layout 4
            ppre = ""  # remove
            ppost = ""  # remove
        else:
            assert False
    elif mode == MM.BLOCKEND:
        if not ppre and not ppost:  # Layout 1
            pass  # warn at BLOCKSTART
        elif not ppre and ppost == '</p>':  # Layout 2
            ppre = ppost  # shift </p> from back to front
            ppost = ""
        elif ppre == '<p>' and not ppost:  # Layout 3
            ppre = ""  # remove
        elif ppre == '<p>' and ppost == '</p>':  # Layout 4
            ppre = ""  # remove
            ppost = ""  # remove
        else:
            assert False, mm.group()
    else:
        assert False    
    # return f"pre<{ppre}>exp<{expansion}>post<{ppost}>"
    result = f"{ppre}{expansion}{ppost}"
    return result


def get_state(namespace: str) -> tg.Any:
    return macrostate[namespace]  # must exist


def set_state(namespace: str, obj: tg.Any):
    global macrostate
    macrostate[namespace] = obj


def switch_part(newpartname: str):
    """Patch all macrodefs to mention a different part than previously (yes, it's ugly)."""
    for macroname, macrodef in itertools.chain(macrodefs_early.items(), macrodefs_late.items()):
        nargs, mode, expander, switcher = macrodef
        switcher(macroname, newpartname)


def register_macro(name: str, numargs: int, mode: MM, expander: Macroexpander, 
                   switcher: Partswitcher = lambda mn, pn: None, 
                   redefine=False):
    global macrodefs_early, macrodefs_late
    macrodefs = (macrodefs_early if mode == MM.EARLY else macrodefs_late)
    if redefine:
        assert name in macrodefs
    else:
        assert name not in macrodefs
    assert 0 <= numargs <= 2
    assert name == name.upper()  # macronames must be all uppercase
    macrodefs[name] = (numargs, mode, expander, switcher)


def _testmode_reset():
    global macrodefs_early, macrodefs_late, macrostate
    macrodefs_early, macrodefs_late = (dict(), dict())
    macrostate = dict()


from blessed import Terminal
import typing as tg
import webbrowser

import base as b
import sdrl.constants as c
import sdrl.repo as r


def prefix(entry: r.ReportEntry, selected: dict[str, bool], rejected: dict[str, bool]):
    if selected[entry[0]]:
        return c.INTERACT_ACCEPT_SYMBOL
    if not(rejected is None) and rejected[entry[0]]:
        return c.INTERACT_REJECT_SYMBOL
    return " "


def redraw_filter_selection(term: Terminal, entries: tg.Sequence[r.ReportEntry], rowindex: int, 
                            selected: dict[str, bool], rejected: dict[str, bool], course_url: str):
    print(term.home + term.clear, end="")
    lines = term.height - 1
    print("Move with arrow keys, select/deselect with space/enter, exit with 'Q''")
    if not(course_url is None):
        print("Press 'o' to open task in browser")
        lines = lines - 1
    rangestart = max(0, rowindex - lines // 2)
    rangeend = min(len(entries), rowindex + lines // 2)
    for i in range(rangestart, rangeend):
        if i == rowindex:
            print(term.reverse, end="")
        else:
            print(term.normal, end="")
        print(prefix(entries[i], selected, rejected) + " %4.2fh " % entries[i][1] + entries[i][0], end="")
        print(" " * (term.width - 9 - len(entries[i][0])))  # padding to end of line
    print(term.normal, end="")  # if we end on the selection


def filter_entries(entries: tg.Sequence[r.ReportEntry], selected: dict[str, bool], 
                   rejected: dict[str, bool] | None, course_url: str | None):
    term = Terminal()
    rowindex = 0
    with term.cbreak(), term.hidden_cursor():
        inp = None
        while not inp or not (str(inp) in "Qq" or str(inp.name) == "KEY_ESCAPE"):
            redraw_filter_selection(term, entries, rowindex, selected, rejected, course_url)
            inp = term.inkey()
            if str(inp) == " " or str(inp.name) == "KEY_ENTER":
                if rejected is None:
                    selected[entries[rowindex][0]] = not(selected[entries[rowindex][0]])
                elif rejected[entries[rowindex][0]]:
                    rejected[entries[rowindex][0]] = False
                elif selected[entries[rowindex][0]]:
                    selected[entries[rowindex][0]] = False
                    rejected[entries[rowindex][0]] = True
                else:
                    selected[entries[rowindex][0]] = True
            elif str(inp) == "j" or str(inp.name) == "KEY_DOWN":
                if rowindex < len(entries) - 1:
                    rowindex += 1
            elif str(inp) == "k" or str(inp.name) == "KEY_UP":
                if rowindex > 0:
                    rowindex -= 1
            elif not(course_url is None) and (str(inp) == "o" or str(inp) == "O"):
                webbrowser.open(f"{course_url}/{entries[rowindex][0]}.html")  # TODO 2 on WSL, see 
                # https://github.com/python/cpython/issues/89752


def select_entries(entries: tg.Sequence[r.ReportEntry]):
    selected = {entry[0]: True for entry in entries}
    filter_entries(entries, selected, None, None)
    return list(filter(lambda entry: selected[entry[0]], entries))


def grade_entries(entries: list[r.ReportEntry], course_url: str, override: bool, filter_method = None) -> list[str] | None:
    """
    Mark entries in list as accepted if they are and count rejection if not.
    Returns list of new rejections because the count is not enough to know whether they were rejected in current run
    """
    submission = b.slurp_yaml(c.SUBMISSION_FILE)
    selected = {entry[0]: currently_accepted(submission, entry, override) for entry in entries}
    rejected = {entry[0]: currently_rejected(submission, entry, override) for entry in entries}
    (filter_method or filter_entries)(entries, selected, rejected, course_url)
    if not any(selected.values()) and not any(rejected.values()):
        return None  # nothing to do

    def processed(entry: r.ReportEntry) -> r.ReportEntry:
        taskname, workhoursum, timevalue, rejections, accepted = entry
        if selected[taskname]:
            if override and currently_accepted(submission, entry, override):
                selected[taskname] = False
                return None
            accepted = True
        if rejected[taskname]:
            if override and currently_rejected(submission, entry, override):
                rejected[taskname] = False
                return None
            rejections += 1
        return taskname, workhoursum, timevalue, rejections, accepted

    for i in reversed(range(len(entries))):
        entries[i] = processed(entries[i])
        if not entries[i]:
            del entries[i]
    return [key for key, value in rejected.items() if value]

def currently_accepted(submission: b.StrAnyDict, entry: r.ReportEntry, override: bool) -> bool:
    return entry[4] if override else r.is_accepted(submission.get(entry[0]) or "")

def currently_rejected(submission: b.StrAnyDict, entry: r.ReportEntry, override: bool) -> bool:
    return not(currently_accepted(submission, entry, override)) if override else r.is_rejected(submission.get(entry[0]) or "")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'py',
 'sdrl': 'py/sdrl',
 'sdrl.subcmd': 'py/sdrl/subcmd',
 'sdrl.tests': 'py/sdrl/tests',
 'subcmd': 'py/subcmd'}

packages = \
['py',
 'py.sdrl',
 'py.sdrl.subcmd',
 'py.sdrl.tests',
 'py.tests',
 'sdrl',
 'sdrl.subcmd',
 'sdrl.tests',
 'subcmd']

package_data = \
{'': ['*'],
 'py.sdrl.tests': ['data/*',
                   'data/sedrila-test1/*',
                   'data/sedrila-test1/.git/*',
                   'data/sedrila-test1/.git/hooks/*',
                   'data/sedrila-test1/.git/info/*',
                   'data/sedrila-test1/.git/logs/*',
                   'data/sedrila-test1/.git/logs/refs/heads/*',
                   'data/sedrila-test1/.git/logs/refs/remotes/origin/*',
                   'data/sedrila-test1/.git/objects/04/*',
                   'data/sedrila-test1/.git/objects/08/*',
                   'data/sedrila-test1/.git/objects/0d/*',
                   'data/sedrila-test1/.git/objects/0e/*',
                   'data/sedrila-test1/.git/objects/17/*',
                   'data/sedrila-test1/.git/objects/22/*',
                   'data/sedrila-test1/.git/objects/23/*',
                   'data/sedrila-test1/.git/objects/29/*',
                   'data/sedrila-test1/.git/objects/30/*',
                   'data/sedrila-test1/.git/objects/35/*',
                   'data/sedrila-test1/.git/objects/42/*',
                   'data/sedrila-test1/.git/objects/4f/*',
                   'data/sedrila-test1/.git/objects/5e/*',
                   'data/sedrila-test1/.git/objects/64/*',
                   'data/sedrila-test1/.git/objects/65/*',
                   'data/sedrila-test1/.git/objects/70/*',
                   'data/sedrila-test1/.git/objects/7a/*',
                   'data/sedrila-test1/.git/objects/7b/*',
                   'data/sedrila-test1/.git/objects/82/*',
                   'data/sedrila-test1/.git/objects/87/*',
                   'data/sedrila-test1/.git/objects/93/*',
                   'data/sedrila-test1/.git/objects/9d/*',
                   'data/sedrila-test1/.git/objects/a7/*',
                   'data/sedrila-test1/.git/objects/a9/*',
                   'data/sedrila-test1/.git/objects/b4/*',
                   'data/sedrila-test1/.git/objects/bf/*',
                   'data/sedrila-test1/.git/objects/c1/*',
                   'data/sedrila-test1/.git/objects/c2/*',
                   'data/sedrila-test1/.git/objects/c4/*',
                   'data/sedrila-test1/.git/objects/cd/*',
                   'data/sedrila-test1/.git/objects/ce/*',
                   'data/sedrila-test1/.git/objects/d6/*',
                   'data/sedrila-test1/.git/objects/e0/*',
                   'data/sedrila-test1/.git/objects/e2/*',
                   'data/sedrila-test1/.git/objects/e3/*',
                   'data/sedrila-test1/.git/objects/f2/*',
                   'data/sedrila-test1/.git/objects/pack/*',
                   'data/sedrila-test1/.git/refs/heads/*',
                   'data/sedrila-test1/.git/refs/remotes/origin/*',
                   'data/sedrila-test1/task/*',
                   'input/*',
                   'input/ch/*',
                   'input/ch/ch1/*',
                   'input/ch/ch1/myarchive.zip/*',
                   'input/ch/ch1/tg11/*',
                   'input/ch/ch1/tg12/*',
                   'input/ch/ch1/tg13draft/*',
                   'input/ch/ch2draft/*',
                   'input/ch/ch2draft/tg21/*',
                   'input/itree.zip/*'],
 'sdrl.tests': ['data/*',
                'data/sedrila-test1/*',
                'data/sedrila-test1/.git/*',
                'data/sedrila-test1/.git/hooks/*',
                'data/sedrila-test1/.git/info/*',
                'data/sedrila-test1/.git/logs/*',
                'data/sedrila-test1/.git/logs/refs/heads/*',
                'data/sedrila-test1/.git/logs/refs/remotes/origin/*',
                'data/sedrila-test1/.git/objects/04/*',
                'data/sedrila-test1/.git/objects/08/*',
                'data/sedrila-test1/.git/objects/0d/*',
                'data/sedrila-test1/.git/objects/0e/*',
                'data/sedrila-test1/.git/objects/17/*',
                'data/sedrila-test1/.git/objects/22/*',
                'data/sedrila-test1/.git/objects/23/*',
                'data/sedrila-test1/.git/objects/29/*',
                'data/sedrila-test1/.git/objects/30/*',
                'data/sedrila-test1/.git/objects/35/*',
                'data/sedrila-test1/.git/objects/42/*',
                'data/sedrila-test1/.git/objects/4f/*',
                'data/sedrila-test1/.git/objects/5e/*',
                'data/sedrila-test1/.git/objects/64/*',
                'data/sedrila-test1/.git/objects/65/*',
                'data/sedrila-test1/.git/objects/70/*',
                'data/sedrila-test1/.git/objects/7a/*',
                'data/sedrila-test1/.git/objects/7b/*',
                'data/sedrila-test1/.git/objects/82/*',
                'data/sedrila-test1/.git/objects/87/*',
                'data/sedrila-test1/.git/objects/93/*',
                'data/sedrila-test1/.git/objects/9d/*',
                'data/sedrila-test1/.git/objects/a7/*',
                'data/sedrila-test1/.git/objects/a9/*',
                'data/sedrila-test1/.git/objects/b4/*',
                'data/sedrila-test1/.git/objects/bf/*',
                'data/sedrila-test1/.git/objects/c1/*',
                'data/sedrila-test1/.git/objects/c2/*',
                'data/sedrila-test1/.git/objects/c4/*',
                'data/sedrila-test1/.git/objects/cd/*',
                'data/sedrila-test1/.git/objects/ce/*',
                'data/sedrila-test1/.git/objects/d6/*',
                'data/sedrila-test1/.git/objects/e0/*',
                'data/sedrila-test1/.git/objects/e2/*',
                'data/sedrila-test1/.git/objects/e3/*',
                'data/sedrila-test1/.git/objects/f2/*',
                'data/sedrila-test1/.git/objects/pack/*',
                'data/sedrila-test1/.git/refs/heads/*',
                'data/sedrila-test1/.git/refs/remotes/origin/*',
                'data/sedrila-test1/task/*',
                'input/*',
                'input/ch/*',
                'input/ch/ch1/*',
                'input/ch/ch1/myarchive.zip/*',
                'input/ch/ch1/tg11/*',
                'input/ch/ch1/tg12/*',
                'input/ch/ch1/tg13draft/*',
                'input/ch/ch2draft/*',
                'input/ch/ch2draft/tg21/*',
                'input/itree.zip/*']}

install_requires = \
['Jinja2>=3.1,<4.0',
 'Markdown>=3.7,<4.0',
 'PyYAML>=6.0,<7.0',
 'Pygments>=2.17,<3.0',
 'argparse_subcommand>=1.2,<2.0',
 'blessed>=1.20,<2.0',
 'mdx_linkify>=2.1,<3.0',
 'requests>=2.32,<3.0',
 'rich>=13.7,<14.0']

entry_points = \
{'console_scripts': ['sedrila = sedrila:main']}

setup_kwargs = {
    'name': 'sedrila',
    'version': '2.3.0',
    'description': 'Tool infrastructure for building and running "self-driven lab" courses',
    'long_description': '[![Documentation Status](https://readthedocs.org/projects/sedrila/badge/?version=latest)](https://sedrila.readthedocs.io/en/latest/?badge=latest)\n\n# `sedrila`: Tool infrastructure for building and running "self-driven lab" courses\n\nA "self-driven lab" (SeDriLa) course is one where students select freely \na subset from a large set of tasks.\nThe tasks are described with sufficient detail that no guidance from an instructor\nis needed most of the time.\n\nsedrila is a command-line tool supporting course authors for authoring a course\nand then course instructors and students for executing it.\n\nFind the [documentation at readthedocs](https://sedrila.readthedocs.io).\n\n\n## 1. Ideas for future versions\n\n\n### 1.1 A currently needed refactoring: Target directory structure\n\nThe current layout of the source tree is wrong.\nCurrently, the `templates` and `baseresources` directories will end up \nas top-level directories when the package is installed,\nwhich means they will clash with any top-level modules of that name\nanywhere in our dependencies.\n\nWe need to perform the following refactorings to arrive at a proper structure:\n\n- `py` --> `sedrila`: This will be the top level directory that gets installed.\n- `sedrila/sdrl/*` --> `sedrila/*`: We remove the now-intermediate namespace.\n  This implies joining the current `sdrl/tests` into `sedrila/tests`.\n- `templates` --> `sedrila/templates`: The HTML templates simply become part of the\n  tree to be installed.\n- `baseresources` --> `sedrila/baseresources`: Ditto.\n\nThese changes require a lot of changes of import statements.\nFor instance, the current module `base` will become `sedrila.base`\nand `sdrl.course` will become `sedrila.course`.\nThe logic for computing `sedrila_libdir` in `courses.py` must be adapted.\n`SedrilaArgParser.get_version()` must be adapted.\nThe files lists in `pyproject.toml` must be corrected.\n\n\n### 1.2 `instructor`: Handling instructors\' trees of student repos\n\n- Process `SEDRILA_INSTRUCTOR_COURSE_URLS` as described in the instructor documentation.\n- `sedrila instructor` should keep a JSON file `student_course_urls.json` that maps student usernames\n  to the course URL first seen for that student, because if a student ever changed\n  the URL in the `student.yaml`, prior signed commits of instructors might become \n  invalid semantically if the new course has a different set of tasks.  \n  The map is added to when a `student.yaml` is first seen\n  and checked against at each later time.  \n  Note that a student taking part a second time, with a fresh repo,\n  might require manual editing of that JSON file to remove that entry.\n- Better yet, there could be an option `sedrial instructor --allow-repo2` that \n  performs that editing automatically\n  and also checks that the new repo contains no instructor-signed commits.\n- Command `sedrila instructor --clean-up-repos-home`\n  to clean up instructor work directory trees-of-trees\n  by deleting all level-1 subtrees in which the `student.yaml`\n  has a `course_url` that is not mentioned in the \n  `SEDRILA_INSTRUCTOR_COURSE_URLS`environment variable.\n  This option should ask a safety question before starting to work.\n- Add `sedrila instructor --http` which presents the local directory tree to localhost as follows:\n    - Show directory, each file is a hyperlink, including `..` (except in the starting directory)\n    - *.md files get rendered as Markdown\n    - *.txt files get shown verbatim\n    - *.py file contents are Markdown-rendered as a Python code block. Ditto for other languages.\n\n\n## 2. Development process: TODO-handling during development\n\nWe use this convention for the development of `sedrila`.\nIt may also be helpful for course authors if the team is small enough.\n\nIf something is incomplete, add a TODO marker with a priorization digit and\nadd a short description of what needs to be done. Examples:\n- `TODO 1: find proper formulation`\n- `TODO 2: restructure to use ACME lib`\n- `TODO 3: add automatic grammar correction`\n\nPriorities:\n- 1: to be completed soon (within a few days)\n- 2: to be completed once the prio 1 things are done (within days or a few weeks)\n- 3: to be completed at some later time (usually several weeks or more into the future,\n  because it is big) or never (because it is not-so-important: "nice-to-have features")\n\nThen use the IDE global search to work through these layer-by-layer.\nDemote items to a lower priority when they become stale or remove them.\nKick out prio 3 items when they become unlikely.\n',
    'author': 'Lutz Prechelt',
    'author_email': 'prechelt@inf.fu-berlin.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fubinf/sedrila',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



"""Helper functionality extracted from Parts and Pieces to make them lighter."""
import glob
import os
import re
import typing as tg

import jinja2
import yaml

import base as b
import cache
import sdrl.directory as dir
import sdrl.elements as el
import sdrl.html as h


class PartbuilderMixin:  # to be mixed into a Part class
    TEMPLATENAME = "??.html"  # defined by each partbuilder class

    directory: dir.Directory
    metadata_text: str  # the YAML front matter character stream
    metadata: b.StrAnyDict  # the YAML front matter
    content: str  # the markdown block
    stage: str = ''  # stage: value
    skipthis: bool  # do not include this chapter/taskgroup/task in generated site
    toc: str  # table of contents

    @property
    def breadcrumb_item(self) -> str:
        titleattr = f"title=\"{h.as_attribute(self.title)}\""
        return f"<a href='{self.outputfile}' {titleattr}>{self.name}</a>"

    @property
    def to_be_skipped(self) -> bool:
        return False  # redefined in concrete part classes

    @property
    def toc_entry(self) -> str:
        classes = f"stage-{self.stage}" if self.stage else "no-stage"
        return h.indented_block(self.toc_link_text, self.TOC_LEVEL, classes)  # noqa

    @property
    def toc_link_text(self) -> str:
        titleattr = f"title=\"{self.title}\""  # noqa
        return f"<a href='{self.outputfile}' {titleattr}>{self.name}</a>"  # noqa

    def do_build(self):
        b.debug(f"do_build({self.cache_key}): skip={self.to_be_skipped}")
        if self.to_be_skipped:
            return
        body_s = self.directory.get_the(el.Body_s, self.name).value  # noqa
        body_i = self.directory.get_the(el.Body_i, self.name).value  # noqa
        # if only the [INSTRUCTOR] part content has changed, we need not build the student file:
        changed_deps = [dep for dep in self.my_dependencies() if dep.state == cache.State.HAS_CHANGED]
        b.debug(str([dep.cache_key for dep in changed_deps]))
        if len(changed_deps) == 1 and isinstance(changed_deps[0], el.Body_i):
            self.render_structure(self.course, self, body_i, self.targetdir_i)  # noqa
        else:
            self.render_structure(self.course, self, body_s, self.targetdir_s)  # noqa
            self.render_structure(self.course, self, body_i, self.targetdir_i, info=False)  # noqa

    def as_json(self) -> b.StrAnyDict:
        return dict(title=self.title)  # noqa

    def evaluate_stage(self, context: str, course) -> None:
        """
        Cut the 'stage' attribute down to its first word, check it against course.stages, report violations.
        Set self.skipthis according to course.include_stage.
        """
        self.skipthis = False  # default case
        # ----- handle parts with no 'stage:' given:
        if not getattr(self, 'stage', ''):
            return
        # ----- extract first word from stage:
        mm = re.match(r'\w+', self.stage)  # match first word
        stageword = mm.group(0) if mm else ''
        # ----- handle parts with unknown stage (error):
        try:
            stage_index = course.stages.index(stageword)
        except ValueError:
            b.error(f"Illegal value of 'stage': '{stageword}'", file=context)
            return
        # ----- handle parts with known stage:
        self.skipthis = course.include_stage_index > stage_index

    def find_zipdirs(self):
        """find all dirs (not files!) *.zip in self.sourcefile dir (not below!), warns about *.zip files"""
        inputdir = os.path.dirname(self.sourcefile)  # noqa
        zipdirs = glob.glob(f"{inputdir}/*.zip")
        for zipdirname in zipdirs:
            if os.path.isdir(zipdirname):
                zipfilename = os.path.basename(zipdirname)
                self.directory.make_the(el.Zipdir, zipdirname)
                self.directory.make_the(el.Zipfile, zipfilename, parent=self.course, 
                                        sourcefile=zipdirname, title=zipfilename)
            else:
                b.warning(f"'{zipdirname}' is a file, not a dir, and will be ignored.", file=self.sourcefile)
    
    def make_std_dependencies(self, use_toc_of: el.Part):
        """Create direct and indirect dependencies of Course, Chapter, Taskgroup, and Task Parts."""
        # ----- indirect dependency:
        self.directory.make_the(el.Sourcefile, self.sourcefile, part=self)
        # ----- direct dependencies, who know and create further direct dependencies:
        self.make_dependency(el.Topmatter, part=self)
        self.make_dependency(el.Body_s, part=self, includelist_class=el.IncludeList_s)
        self.make_dependency(el.Body_i, part=self, includelist_class=el.IncludeList_i)
        self.make_dependency(el.TermrefList, part=self)
        self.make_or_get_dependency(el.Toc, name=use_toc_of.name, part=use_toc_of)

    def process_topmatter(self, sourcefile: str, topmatter: b.StrAnyDict, course):
        assert False  # must be defined in concrete classes

    def read_partsfile(self, file: str):
        """
        Reads files consisting of YAML metadata, then Markdown text, separated by a tiple-dash line.
        Stores metadata into self.metadata, rest into self.content.
        """
        SEPARATOR = "---\n"
        # ----- obtain file contents:
        text = b.slurp(file)
        if SEPARATOR not in text:
            b.error(f"triple-dash separator is missing", file=self.sourcefile)  # noqa
            return
        self.metadata_text, self.content = text.split(SEPARATOR, 1)
        # ----- parse metadata
        try:
            # ----- parse YAML data:
            self.metadata = yaml.safe_load(self.metadata_text)
        except yaml.YAMLError as exc:
            b.error(f"metadata YAML is malformed: {str(exc)}", file=self.sourcefile)  # noqa
            self.metadata = dict()  # use empty metadata as a weak replacement

    def render_structure(self, course, part: el.Part, body: str, targetdir: str, info=True):
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(course.templatedir), autoescape=False)
        template = env.get_template(self.TEMPLATENAME)  # noqa
        output = template.render(sitetitle=course.title,
                                 breadcrumb=h.breadcrumb(*self.structure_path()[::-1]),
                                 title=part.title,
                                 linkslist_top=getattr(part, 'linkslist_top', ""),
                                 linkslist_bottom=getattr(part, 'linkslist_bottom', ""),
                                 part=self,
                                 toc=part.toc,
                                 content=body)
        b.spit(f"{targetdir}/{self.outputfile}", output)  # noqa
        if info:
            b.info(f"{targetdir}/{self.outputfile}")  # noqa

    def structure_path(structure: el.Part) -> list[el.Part]:
        """List of nested parts, from a given part up to the course."""
        import sdrl.course
        path = []
        if isinstance(structure, sdrl.course.Task):
            path.append(structure)
            structure = structure.taskgroup
        if isinstance(structure, sdrl.course.Taskgroup):
            path.append(structure)
            structure = structure.chapter
        if isinstance(structure, sdrl.course.Chapter):
            path.append(structure)
            structure = structure.course
        if isinstance(structure, sdrl.course.Course):
            path.append(structure)
        return path

    def _make(self, mytype, **kwargs):  # abbrev
        self.add_dependency(self.directory.make_the(mytype, self.name, **kwargs))  # noqa


def toc(structure: el.Part) -> str:
    """Return a table-of-contents HTML fragment for the given structure via structural recursion."""
    import sdrl.course
    parts = structure.structure_path()
    fulltoc = len(parts) == 1  # path only contains course
    assert isinstance(parts[-1], sdrl.course.Coursebuilder)
    course = tg.cast(sdrl.course.Coursebuilder, parts[-1])
    result = ['']  # start with a newline
    for chapter in course.chapters:  # noqa
        if chapter.to_be_skipped:
            continue
        result.append(chapter.toc_entry)
        if not fulltoc and chapter not in parts:
            continue
        for taskgroup in chapter.taskgroups:
            effective_tasklist = [t for t in course.taskorder 
                                  if t in taskgroup.tasks and not t.to_be_skipped]
            if taskgroup.to_be_skipped or not effective_tasklist:
                continue
            result.append(taskgroup.toc_entry)
            if not fulltoc and taskgroup not in parts:
                continue
            for task in effective_tasklist:
                result.append(task.toc_entry)
    result.append(course.glossary.toc_entry)
    return "\n".join(result)



"""
Logic for handling information coming from git repos: 
effort commits, submissions, submission checks.
"""
import re
import subprocess as sp
import typing as tg

import yaml

import base as b
import git
import sdrl.constants as c
import sdrl.course


CheckedTuple = tg.Tuple[str, str, str]  # hash, taskname, tasknote
ReportEntry = tg.Tuple[str, float, float, int, bool]  # taskname, workhoursum, timevalue, rejections, accepted


def import_gpg_keys(instructors: tg.Sequence[b.StrAnyDict]):
    for instructor in instructors:
        if not(instructor.get('pubkey')):
            b.warning("No key present for " + instructor.get('nameish') + ", skipping")
            continue
        b.info("Importing key for " + instructor.get('nameish'))
        if type(instructor['pubkey']) is list:
            instructor['pubkey'] = "\n".join(instructor['pubkey'])
        if not instructor['pubkey'].startswith("-----"):
            instructor['pubkey'] = ("-----BEGIN PGP PUBLIC KEY BLOCK-----\n"
                                    f"\n{instructor['pubkey']}\n"
                                    "-----END PGP PUBLIC KEY BLOCK-----")
        sp.run(["gpg", "--import"], input=instructor['pubkey'], encoding='ascii')


def compute_student_work_so_far(course: sdrl.course.Course, commits: tg.Sequence[git.Commit]):
    """
    Obtain per-task worktimes from student commits and 
    per-task timevalues from submission checked commits.
    Store them in course.
    """
    _accumulate_workhours_per_task(commits, course)
    hasheslist = _submission_checked_commit_hashes(course, commits)
    checked_tuples = _checked_tuples_from_commits(hasheslist, course)
    _accumulate_timevalues_and_attempts(checked_tuples, course)


def student_work_so_far(course) -> tg.Tuple[list[ReportEntry], float, float]:
    """Data for work report: retrieve pre-computed per-task entries (worktime, attempts) and totals."""
    workhours_total = 0.0
    timevalue_total = 0.0
    result = []
    for taskname in sorted((t.name for t in course.taskdict.values())):
        task = course.taskdict[taskname]
        if task.workhours != 0.0:
            workhours_total += task.workhours
            if task.is_accepted:
                timevalue_total += task.timevalue
            result.append((taskname, task.workhours, task.timevalue,
                           task.rejections, task.is_accepted))  # one result tuple
    return result, workhours_total, timevalue_total  # overall result triple


def is_accepted(tasknote: str):
    overridden = tasknote.startswith(c.SUBMISSION_OVERRIDE_PREFIX)
    if overridden:
        tasknote = tasknote[len(c.SUBMISSION_OVERRIDE_PREFIX):]
    return tasknote.startswith(c.SUBMISSION_ACCEPT_MARK)


def is_rejected(tasknote: str):
    overridden = tasknote.startswith(c.SUBMISSION_OVERRIDE_PREFIX)
    if overridden:
        tasknote = tasknote[len(c.SUBMISSION_OVERRIDE_PREFIX):]
    return tasknote.startswith(c.SUBMISSION_REJECT_MARK)


def submission_file_entries(entries: tg.Iterable[ReportEntry], rejected: list[str] = None, 
                            override: bool = False) -> dict[str, str]:
    """taskname -> CHECK_MARK  for each yet-to-be-accepted task with effort in any commit."""
    candidates = dict()
    for taskname, workhoursum, timevalue, rejections, accepted in entries:
        if rejected is None:
            if not accepted:
                candidates[taskname] = c.SUBMISSION_CHECK_MARK
        else:
            if taskname in rejected:
                candidates[taskname] = (c.SUBMISSION_OVERRIDE_PREFIX if override else "") + c.SUBMISSION_REJECT_MARK
            elif accepted:
                candidates[taskname] = (c.SUBMISSION_OVERRIDE_PREFIX if override else "") + c.SUBMISSION_ACCEPT_MARK
            else:
                candidates[taskname] = c.SUBMISSION_CHECK_MARK
    return candidates


def _accumulate_timevalues_and_attempts(checked_tuple_groups: tg.Sequence[tg.Sequence[CheckedTuple]],
                                        course: sdrl.course.Course):
    """Reflect the check_tuples data in the course data structure."""
    for checked_tuples in checked_tuple_groups:
        for refid, taskname, tasknote in checked_tuples:
            b.debug(f"tuple: {refid}, {taskname}, {tasknote}")
            task = course.task(taskname)
            requirements = {requirement: course.task(requirement) for requirement in task.requires
                            if course.task(requirement) is not None}  # ignore Taskgroups
            b.debug(f"requirements({task.name}): {requirements}")
            open_requirements = [taskname for taskname, task in requirements.items()
                                 if not(any(taskname == tname for (_, tname, _) in checked_tuples)) and
                                 not task.is_accepted and task.remaining_attempts]
            if open_requirements:
                b.warning(f"Attempted to grade task {taskname} with missing requirements: {open_requirements}")
                continue
            overridden = tasknote.startswith(c.SUBMISSION_OVERRIDE_PREFIX)
            if overridden:
                tasknote = tasknote[len(c.SUBMISSION_OVERRIDE_PREFIX):]
            if tasknote.startswith(c.SUBMISSION_ACCEPT_MARK):
                if not task.remaining_attempts and not overridden:
                    b.warning(f"Cannot accept task that has consumed its allowed attempts: {taskname}")
                else:
                    if overridden and task.rejections:
                        task.rejections -= 1
                    task.is_accepted = True
                    b.debug(f"{taskname} accepted, {'' if overridden else 'not '} overridden")
            elif tasknote.startswith(c.SUBMISSION_REJECT_MARK):
                if not task.is_accepted or overridden:
                    if overridden:
                        task.is_accepted = False
                    task.rejections += 1
                    b.debug(f"{taskname} rejected, {'' if overridden else 'not '} overridden")
            else:
                pass  # unmodified entry: instructor has not checked it


def _accumulate_workhours_per_task(commits: tg.Iterable[git.Commit], course: sdrl.course.Course):
    """Reflect the workentries data in the course data structure."""
    for commit in commits:
        parts = _parse_taskname_workhours(commit.subject)
        if parts is None:  # this is no worktime commit
            continue
        taskname, worktime = parts  # unpack
        if taskname in course.taskdict:
            task = course.taskdict[taskname]
            task.workhours += worktime
        else:
            b.warning(f"Commit '{commit.subject}': Task '{taskname}' does not exist. Entry ignored.")


def _checked_tuples_from_commits(hasheslist: list[list[str]], course: sdrl.course.Course
                                 ) -> tg.Sequence[tg.Sequence[CheckedTuple]]:
    """Collect the individual checks across all 'submission.yaml checked' commits.
    This will only allow grades for tasks that were actually requested to grade.
    The state of a task should always be the last one in a given group."""
    result = []
    for hashes in hasheslist:
        if not hashes:
            continue
        try:
            gitfile = git.contents_of_file_version(hashes.pop(0), c.SUBMISSION_FILE, encoding='utf8')
            requested = list(yaml.safe_load(gitfile).keys())
        except Exception:  # noqa
            continue  # no submission file yet, grading isn't possible
        groupdict: dict[str, CheckedTuple] = {}
        for refid in hashes:
            submission = yaml.safe_load(git.contents_of_file_version(refid, c.SUBMISSION_FILE, encoding='utf8'))
            for taskname, tasknote in submission.items():
                if not(taskname in requested):
                    b.warning(f"Attempted to grade task that wasn't requested: {taskname}")
                    continue
                groupdict[taskname] = (refid, taskname, tasknote)
        if groupdict:
            result.append(groupdict.values())
    return result


def _submission_checked_commit_hashes(course: sdrl.course.Course,
                                      commits: tg.Sequence[git.Commit]) -> list[list[str]]:
    """List of hashes of properly instructor-signed commits of finished submission checks.
    This will be grouped by consecutively done commits of instructors and student, starting with the grading request."""
    try:
        allowed_signers = [b.as_fingerprint(instructor['keyfingerprint'])
                           for instructor in course.instructors]
    except KeyError:
        b.critical("missing 'keyfingerprint' in configuration")
    group = []
    result = [group]
    student_commit = None
    for commit in commits:
        b.debug(f"commit.subject, fpr: {commit.subject}, {commit.key_fingerprint}")
        right_subject = re.match(c.SUBMISSION_CHECKED_COMMIT_MSG, commit.subject) is not None
        right_signer = commit.key_fingerprint and b.as_fingerprint(commit.key_fingerprint) in allowed_signers  # noqa
        if not right_signer:
            student_commit = commit
            if group:
                group = []
                result.append(group)
        if right_subject and right_signer and student_commit:
            if not group:
                group.append(student_commit.hash)
            group.append(commit.hash)
    return result


def _parse_taskname_workhours(commit_msg: str) -> tg.Optional[tg.Tuple[str, float]]:  # taskname, workhours
    """Return pair of (taskname, workhours) from commit message if present, or None otherwise."""
    worktime_regexp = (r"\s*[#%](?P<name>[\w.-]+?)\s+"
                       r"(?:(?P<dectime>-?\d+(\.\d+)?)|"
                       r"(?P<hh>-?\d+):(?P<mm>\d\d)) ?h\b")  # %MyTask117 3.5h   %Some-Stuff 3:45h
    mm = re.match(worktime_regexp, commit_msg)
    if not mm:
        return None  # not the format we're looking for
    taskname = mm.group('name')
    if mm.group('dectime'):  # decimal time
        workhours = float(mm.group('dectime'))
    else:
        workhours = float(mm.group('hh')) + float(mm.group('mm')) / 60  # hh:mm format
    return taskname, workhours
